with Ada.Text_IO; use Ada.Text_IO;

procedure Hello is
    package heap is
        type max_heap is tagged private;
        function empty(self:in out max_heap) return Boolean;
        function size(self:in out max_heap) return Integer;
        procedure push(self:in out max_heap; k:Integer);
        procedure pop(self:in out max_heap);
        function top(self:in out max_heap) return Integer;
        
        -- for debug
        procedure show(self:in out max_heap);
        
    private
        type heap_array is array(Integer range <>) of Integer;
        type array_pointer is access heap_array;
        type max_heap is tagged record
            size:Integer:=0;
            max_size:Integer:=100;
            data:array_pointer:=new heap_array(1 .. 100);
        end record;
        
        procedure swim(self:in out max_heap; pos:Integer);
        procedure sink(self:in out max_heap; pos:Integer);
    end heap;
    
    package body heap is
        function empty(self:in out max_heap) return Boolean is
        begin
            return self.size=0; 
        end empty;
        
        function size(self:in out max_heap) return Integer is
        begin
            return self.size;
        end size;
        
        procedure push(self:in out max_heap; k:Integer) is
        begin
            if self.size=self.max_size then
                self.max_size:=self.max_size*2;
                declare
                    tmp:array_pointer;
                begin
                    tmp:=new heap_array(1 .. self.max_size);
                    for i in self.data'Range loop
                        tmp(i):=self.data(i);
                    end loop;
                    self.data:=tmp;
                end;
            end if;
            self.size:=self.size+1;
            self.data(self.size):=k;
            swim(self, self.size-1);
        end push;
        
        procedure pop(self:in out max_heap) is 
        begin
            self.data(1) := self.data(self.size);
            self.size := self.size - 1;
            sink(self, 1);
        end pop;        

        function top(self:in out max_heap) return Integer is
        begin
            return self.data(1);
        end top;
        
        procedure swim(self:in out max_heap; pos:Integer) is
            tmp:Integer;
            half_pos:Integer;
            new_pos:Integer;
        begin
            new_pos := pos;
            half_pos := new_pos / 2;
            while new_pos > 1 loop
                if self.data(half_pos) < self.data(new_pos) then
                    tmp := self.data(half_pos);
                    self.data(half_pos) := self.data(new_pos);
                    self.data(new_pos) := tmp;
                    new_pos := half_pos;
                    half_pos := new_pos / 2;
                else exit;
                end if;
            end loop;
        end swim;
        
        procedure sink(self:in out max_heap; pos:Integer) is
            tmp:Integer;
            next_pos:Integer;
            new_pos:Integer;
        begin
            new_pos := pos;
            next_pos := new_pos * 2;
            while next_pos <= self.size loop
                if next_pos < self.size then
                    if self.data(next_pos) < self.data(next_pos+1) then
                        next_pos := next_pos + 1;
                    end if;
                end if;
                
                if self.data(new_pos) < self.data(next_pos) then
                    tmp := self.data(new_pos);
                    self.data(new_pos) := self.data(next_pos);
                    self.data(next_pos) := tmp;
                    new_pos := next_pos;
                    next_pos := new_pos * 2;
                else exit;
                end if;
            end loop;
        end sink;
        
        procedure show(self:in out max_heap) is
        begin
            for i in self.data'First .. self.size loop
                Put(Integer'Image(self.data(i)));
            end loop;
            New_Line;
        end show;
    end heap;
    
    use heap;
    
begin
    declare
        h:max_heap;
    begin
        push(h, 3);
        push(h, 1);
        push(h, 2);
        -- show(h);
        Put_Line("Size: " & Integer'Image(size(h)));
        Put_Line("Top element: " & Integer'Image(top(h)));
        pop(h);
        Put_Line("Top element: " & Integer'Image(top(h)));
        -- show(h);
        Put_Line("Is empty: " & Boolean'Image(empty(h)));
        Put_Line("Size: " & Integer'Image(size(h)));
    end;
end Hello;
